#ifndef _layerexsave_compress_hpp_
#define _layerexsave_compress_hpp_

#include "utils.hpp"

#include <vector>

typedef bool ProgressFunc(int percent, void *userdata);

class CompressBase {
	enum { INITIAL_DATASIZE = 1024*100 };

protected:
	ProgressFunc *progress;
	void         *progressData;

	typedef unsigned char BYTE;
	typedef std::vector<BYTE> DATA;
	DATA data;      //< i[f[^
	ULONG cur;      //< i[ʒu
	ULONG size;     //< i[TCY
	ULONG dataSize; //< f[^̈mۃTCY

public:
	/**
	 * RXgN^
	 */
	CompressBase(ProgressFunc *_progress=NULL, void *_progressData=NULL)
		: progress(_progress), progressData(_progressData),
		  cur(0), size(0), dataSize(INITIAL_DATASIZE)
	{
		data.resize(dataSize);
	}
	CompressBase(CompressBase const *ref)
		: progress(ref->progress), progressData(ref->progressData),
		  cur(0), size(0), dataSize(INITIAL_DATASIZE)
	{
		data.resize(dataSize);
	}

	/**
	 * fXgN^
	 */
	virtual ~CompressBase() {}

	/**
	 * vOX
	 * @return LZꂽ
	 */
	bool doProgress(int percent) {
		return (progress && progress(percent, progressData));
	}

	/**
	 * TCYύX
	 * wʒu͂邾̃TCYmۂB
	 * w肵őTCYێB
	 * @param TCY
	 */
	inline void resize(size_t s) {
		if (s > size) {
			size = s;
			if (size > dataSize) {
				dataSize = size * 2;
				data.resize(dataSize);
			}
		}
	}

	/**
	 * 8bitl̏o
	 * @param num l
	 */
	template <typename ANYINT>
	inline void writeInt8(ANYINT num) {
		resize(cur + 1);
		data[cur++] = num & 0xff;
	}
	
	/**
	 * 32bitl̏o
	 * @param num l
	 */
	template <typename ANYINT>
	inline void writeInt32(ANYINT num) {
		writeInt32(num, cur);
		cur += 4;
	}
	template <typename ANYINT>
	inline void writeBigInt32(ANYINT num) {
		writeBigInt32(num, cur);
		cur += 4;
	}

	/**
	 * 32bitl̏o
	 * @param num l
	 */
	template <typename ANYINT>
	inline void writeInt32(ANYINT num, int cur) {
		resize(cur + 4);
		data[cur++] =  num        & 0xff;
		data[cur++] = (num >> 8)  & 0xff;
		data[cur++] = (num >> 16) & 0xff;
		data[cur++] = (num >> 24) & 0xff;
	}

	/**
	 * 32bitl̏o
	 * @param num l
	 */
	template <typename ANYINT>
	inline void writeBigInt32(ANYINT num, int cur) {
		resize(cur + 4);
		data[cur++] = (num >> 24) & 0xff;
		data[cur++] = (num >> 16) & 0xff;
		data[cur++] = (num >> 8)  & 0xff;
		data[cur++] =  num        & 0xff;
	}

	/**
	 * obt@̏o
	 * @param buf obt@
	 * @param size o̓oCg
	 */
	void writeBuffer(const void *buf, int size) {
		resize(cur + size);
		memcpy((void*)&data[cur], buf, size);
		cur += size;
	}

	/**
	 * f[^t@Cɏo
	 * @param out o͐Xg[
	 */
	void store(IStream *out) {
		ULONG s;
		out->Write(&data[0], size, &s);
	}

	/**
	 * k
	 * @param width 摜
	 * @param height 摜c
	 * @param buffer 摜obt@
	 * @param pitch 摜f[^̃sb`
	 * @param tagsDict ^O
	 * @return LZꂽ true
	 */
	virtual bool compress(long width, long height, BufRefT buffer, long pitch, iTJSDispatch2 *tagsDict) = 0;

	/**
	 * t@Cɕۑ
	 */
	bool save(iTJSDispatch2 *layer, const tjs_char *filename, iTJSDispatch2 *info) {
		// C摜
		BufRefT buffer;
		long width, height, pitch;
		if (!GetLayerBufferAndSize(layer, width, height, buffer, pitch)) {
			ttstr msg = filename;
			msg += L":invalid layer";
			TVPThrowExceptionMessage(msg.c_str());
		}
		bool canceled = compress(width, height, buffer, pitch, info);

		// kLZĂȂ΃t@Cۑ
		if (!canceled) {
			IStream *out = TVPCreateIStream(filename, TJS_BS_WRITE);
			if (!out) {
				ttstr msg = filename;
				msg += L":can't open";
				TVPThrowExceptionMessage(msg.c_str());
			}
			try {
				// i[
				store(out);
			} catch (...) {
				out->Release();
				throw;
			}
			out->Release();
		}

		return canceled;
	}

};

/**
 * Cӂ̌^ŕۑ
 */
template <class COMPRESS>
struct CompressAndSave {
	typedef COMPRESS CompressClass;
	static bool saveLayerImage(iTJSDispatch2 *layer, const tjs_char *filename, iTJSDispatch2 *info, ProgressFunc *progress=NULL, void *progressData=NULL) {
		CompressClass work(progress, progressData);
		return        work.save(layer, filename, info);
	}
};

#endif
